import os

current_dir = os.path.dirname(os.path.abspath(__file__))

savesDir = os.path.join(current_dir, "saves")
assetsDir = os.path.join(current_dir, "assets")
tmpDir = os.path.join(current_dir, "tmp")
songAliasDBPath = os.path.join(current_dir, "db", "songAlias.db")
songDataDBPath = os.path.join(current_dir, "db", "songs.db")

scorePredictorModelPath = os.path.join(current_dir, "model", "score_predictor.pth")



from . import getInfo
from . import utils
from . import database
from . import config
import json
import time
import io
import traceback
import math
import scipy.stats
import numpy
import random
from concurrent.futures import ThreadPoolExecutor

def getPlayerInfo(userProfile, justDATA=False, justHTML=False):
    userdata = getInfo.getPlayerUserData(userProfile)
    save = getInfo.getSave(userProfile)
    
    results = userdata
    results["playerRating"] = round(save[1]["playerRating"], 4)
    results["TotalPlayCount"] = save[1]["playerPlayRecords"]["TotalPlayCount"]
    results["PlayCountI"] = save[1]["playerPlayRecords"]["PlayCountI"]
    results["PlayCountIi"] = save[1]["playerPlayRecords"]["PlayCountIi"]
    results["PlayCountIii"] = save[1]["playerPlayRecords"]["PlayCountIii"]
    results["PlayCountIv"] = save[1]["playerPlayRecords"]["PlayCountIv"]
    results["PlayTotalFc"] = save[1]["playerPlayRecords"]["TotalFc"]
    results["PlayTotalAp"] = save[1]["playerPlayRecords"]["TotalAp"]
    results["PlayTotalApp"] = save[1]["playerPlayRecords"]["TotalApp"]
    results["PlayTotalEx"] = save[1]["playerPlayRecords"]["TotalEx"]
    results["PlayTotalExPlus"] = save[1]["playerPlayRecords"]["TotalExPlus"]
    results["Combo"] = {}
    results["Combo"]["Tap"] = save[1]["playerPlayRecords"]["Tap"]
    results["Combo"]["Slide"] = save[1]["playerPlayRecords"]["Slide"]
    results["Combo"]["Flick"] = save[1]["playerPlayRecords"]["Flick"]
    results["Combo"]["Catch"] = save[1]["playerPlayRecords"]["Catch"]
    results["Combo"]["Rotate"] = save[1]["playerPlayRecords"]["Rotate"]
    results["Combo"]["Status"] = {}
    results["Combo"]["Status"]["Early"] = save[1]["playerPlayRecords"]["Early"]
    results["Combo"]["Status"]["Late"] = save[1]["playerPlayRecords"]["Late"]
    results["Combo"]["Status"]["Good"] = save[1]["playerPlayRecords"]["Good"]
    results["Combo"]["Status"]["Miss"] = save[1]["playerPlayRecords"]["Miss"]
    results["Combo"]["Status"]["Perfect"] = save[1]["playerPlayRecords"]["Perfect"]
    results["Combo"]["Status"]["PerfectPlus"] = save[1]["playerPlayRecords"]["PerfectPlus"]
    
    # 自定义背景 / 头像 / 角色 (这就是特权) (阶级(?))
    if results["playerName"] == "Terlity":
        results["backgroundID"] = "terlity"
        results["avatarID"] = "terlity"
        results["characterID"] = "empty"
    
    if justDATA:
        return results
    with open(f"{config.assetsDir}/html/me_info.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(results, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((900, 900), 1.5, html_data=html, isHTML=True)

def getFriendAndFollow(userProfile, justDATA=False, justHTML=False):
    results = getInfo.getAllFollowDatas(userProfile)
    
    if justDATA:
        return results
    with open(f"{config.assetsDir}/html/friend_info.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(results, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((1400, 1100), 1.5, html_data=html, isHTML=True)

def getFollowPlayerData(userProfile, shortID, justDATA=False, justHTML=False):
    player_data = {}
    for i in getInfo.getAllFollowDatas(userProfile):
        if i["shortID"] == shortID:
            player_data = i
            break
    
    if justDATA:
        return player_data
    with open(f"{config.assetsDir}/html/friend_code.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(player_data, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((900, 900), 1.5, html_data=html, isHTML=True)

def getFriendBest40(userProfile, shortID, justDATA=False, justHTML=False):
    player_data = {}
    for i in getInfo.getAllFollowDatas(userProfile):
        if i["shortID"] == shortID and i["isFriend"]:
            player_data = i
            break
    if player_data == {}:
        return ""
    return getBest40(utils.shortIDToUserProfile(shortID), justDATA, justHTML)

def getBest40(userProfile, justDATA=False, justHTML=False):
    results = getInfo.getSave(userProfile)
    results[0] = sorted(results[0], key=lambda x: x["rating"], reverse=True)[:40]
    
    if justDATA:
        return results
    with open(f"{config.assetsDir}/html/b40.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(results, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((1400, 1310), 1.5, html_data=html, isHTML=True)

def getBest40RatingChart(userProfile):
    data = getBest40(userProfile, justDATA=True)[0]
    ratings = [round(x["rating"], 4) for x in data]
    song_names = [x["songName"] for x in data]
    return utils.draw_bar_chart(ratings, song_names, save_path=f"{config.tmpDir}/{time.time()}.png", title="Best40 Rating曲线分布")

def getSongsStatus(userProfile, songStatus, justDATA=False, justHTML=False):
    # songStatus [NONE  FC  AP  APP  CLEAR  NOTCLEAR]
    songStatus = list({x for x in [x.replace(" ", "") for x in songStatus.split("/")]})
    getAll = getInfo.getSave(userProfile)
    getAll[1]["songStatus"] = '/'.join(map(str, songStatus))
    results = [[], getAll[1]]
    for songData in getAll[0]:
        if songData["songStatus"] in songStatus:
            if not songData in results[0]:
                results[0].append(songData)
        if "CLEAR" in songStatus and songData["isCleared"]:
            if not songData in results[0]:
                results[0].append(songData)
        if "NOTCLEAR" in songStatus and not songData["isCleared"]:
            if not songData in results[0]:
                results[0].append(songData)
    results[0] == sorted(results[0], key=lambda x: x["rating"], reverse=True)
    
    if justDATA:
        return results
    with open(f"{config.assetsDir}/html/songs_status.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(results, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((1400, 1310), 3, html_data=html, isHTML=True)

def getSong(userProfile, song_id, justDATA=False, justHTML=False):
    # ==========================================================
    # 重点关注(大学)
    gaoxiao_211 = False
    if song_id == "211大学":
        gaoxiao_211 = True
        song_id = "abstruse-dilemma"
    # ==========================================================
    
    data = {}
    getAll = getInfo.getSave(userProfile)
    for song_data in getAll[0]:
        if song_data["songID"] == song_id:
            data[song_data["songLevelName"]] = {"songData": song_data,
                                                "artist": database.songData.getSong(song_id)["artist"]}
    results = [
                {
                    "songID": song_id,
                    "songName": data["I"]["songData"]["songName"],
                    "songArtist": data["I"]["artist"],
                    "songData": data
                },
                getAll[1]
            ]
    """
    results = [
        0: {
            "songID": songID,
            "songName": songName,
            "songArtist": songArtist,
            "songData": {
                "I": {
                    "songData": {
                        "score": score,
                        "level": level,
                        "sourceRating": sourceRating,
                        "songStatus": songStatus,
                        "isCleared": isCleared,
                    },
                    "artist": artist
                }, ...
            }
        },
        1: PlayerInfo
    ]
    """
    
    if justDATA:
        return results
    # ==========================================================
    # 重点关注(大学)
    if gaoxiao_211:
        results[0]["songID"] = f"211大学_{random.randint(1, 2)}"
    # ==========================================================
    with open(f"{config.assetsDir}/html/song.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(results, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((1400, 1100), 1.5, html_data=html, isHTML=True)

def getSongAllScore(song_id, needScoreList=False, needZero=False):
    allScores = {"I": [0, 0], "II": [0, 0], "III": [0, 0], "IV": [0, 0], "IV_Alpha": [0, 0]}
    _allScores = {"I": [], "II": [], "III": [], "IV": [], "IV_Alpha": []}
    for objectID in getInfo.getAllUsersObjectID():
        tmp = getSong({"server": "local", "sign": "", "session": "", "objectId": objectID}, song_id, justDATA=True)
        for level in tmp[0]["songData"]:
            score = tmp[0]["songData"][level]["songData"]["score"]
            if score == 0 and not needZero: continue
            if needScoreList:
                _allScores[level].append(score)
                continue
            allScores[level][0] += 1
            allScores[level][1] += score
    return _allScores if needScoreList else  allScores

def getSongVarianceScore(song_id, needScoreList=False): # 获取方差
    allScores = getSongAllScore(song_id, needScoreList=True)
    results = {}
    for level in allScores: # 先用这个, 如果出错就用 levelScore, levelAverage in zip(allScores, getAverage)
        if len(allScores[level]) == 0:
            results[level] = 0
            continue
        results[level] = numpy.var(allScores[level])
    if needScoreList:
        return [allScores, results]
    return results

def getSongAverageScore(song_id): # 获取平均值
    allScores = getSongAllScore(song_id)
    results = {}
    for level in allScores:
        if allScores[level][0] == 0:
            results[level] = 0
        else:
            results[level] = numpy.mean(allScores[level][1], dtype=int)
            if results[level] > 1010000: results[level] = 1010000
    return results

def getSongStdScore(song_id, scores): # 获取标准差
    getVariances = getSongVarianceScore(song_id, needScoreList=True)
    getAverage = getSongAverageScore(song_id)
    result = {}
    for level in scores: #  先用这个, 如果出错就用 scores, levelVariance, levelAverage in zip(scores, getVariances[1], getAverage)
        mean = getAverage[level]
        variance = getVariances[1][level]
        target_variance = (scores[level] - mean) ** 2
        result[level] = math.sqrt((variance * len(scores) + target_variance) / (len(scores) - 1))
    return result

def getSongSKScore(song_id): # 获取偏度
    getScores = getSongAllScore(song_id, needScoreList=True)
    result = {}
    for level in getScores:
        result[level] = scipy.stats.skew(getScores[level])
    return result

def getSongKurScore(song_id): # 获取峰度
    getScores = getSongAllScore(song_id, needScoreList=True)
    result = {}
    for level in getScores:
        result[level] = scipy.stats.kurtosis(getScores[level])
    return result

def getSongRatingRealRangeData(userProfile, songRatingReal, justDATA=False, justHTML=False):
    data = []
    getAll = getInfo.getSave(userProfile)
    songNeed = database.songData.getSongsRatingRealRange(songRatingReal[0], songRatingReal[1])
    
    for song_data in getAll[0]:
        for song in songNeed:
            if song["id"] == song_data["songID"] and song_data["songLevelName"] in song["levels"] and song_data["score"] != 0:
                data.append(song_data)
    
    data = sorted(data, key=lambda x: x["sourceRating"], reverse=True)
    getAll[0] = data

    if justDATA:
        return getAll
    with open(f"{config.assetsDir}/html/song_ratingreal_range.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(getAll, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((1400, 1310), 3, html_data=html, isHTML=True)

def getSongHistory(userProfile, song_id, justDATA=False):
    getInfo.getSave(userProfile)
    histories = []
    for data in utils.sorted_json_files(f"{config.savesDir}/{userProfile['objectId']}/save", reverse=False, needTimestamp=True):
        timestamp, filePath = int(round(float(data[0]), 0)), data[1]
        histories.append((timestamp, getSong({"server": "history", "historyFilePath": filePath, "sign": "", "session": "", "objectId": ""}, song_id, justDATA=True)[0]))
    LevelScore = {}
    for timestamp, data in histories:
        for level in data["songData"]:
            if level not in LevelScore:
                LevelScore[level] = []
            LevelScore[level].append({
                "timestamp": timestamp,
                "songScore": data["songData"][level]["songData"]["score"],
                "songRating": data["songData"][level]["songData"]["sourceRating"],
                "songStatus": data["songData"][level]["songData"]["songStatus"],
                "songIsCleared": data["songData"][level]["songData"]["isCleared"]
            })
    try:
        for level in LevelScore:
            unique_lst = []
            for x in LevelScore[level]:
                for y in unique_lst:
                    if y["songScore"] >= x["songScore"]:
                        break
                else:
                    unique_lst.append(x)
            LevelScore[level] = unique_lst
    except Exception as e:
        traceback.print_exc()
        LevelScore = {}

    if justDATA:
        return LevelScore
    _save_path = f"{config.tmpDir}/{time.time()}"
    save_paths = []
    for level in LevelScore:
        save_paths.append(f"{_save_path}-{level}.png")
        utils.draw_line_datatime(LevelScore[level], save_paths[-1], f"{level.replace('_', ' ')} Score Trend", "Score")
    return save_paths

def getRatingTrend(userProfile, justDATA=False):
    getInfo.getSave(userProfile)
    histories = []
    for data in utils.sorted_json_files(f"{config.savesDir}/{userProfile['objectId']}/save", reverse=False, needTimestamp=True):
        timestamp, filePath = int(round(float(data[0]), 0)), data[1]
        histories.append((timestamp, getBest40({"server": "history", "historyFilePath": filePath, "sign": "", "session": "", "objectId": ""}, justDATA=True)[1]["playerRating"]))
    
    if justDATA:
        return histories
    return utils.draw_line_datatime(histories, f"{config.tmpDir}/{time.time()}.png", "Rating Trend", "Rating")

def getSongFollowerRanking(userProfile, song_id, vicinity=None, rangeNum=4, justDATA=False, justHTML=False):
    result = {"I": [], "II": [], "III": [], "IV": []}
    # 这个getAllFollowDatas只能获取到关注者的数据, 没有自己的分数
    for user in getInfo.getAllFollowDatas(userProfile):
        for level in result:
            result[level].append({
                "playerName": user["playerName"],
                "avatarID": user["avatarID"],
                "songScore": user["songScores"][song_id][level.lower()],
                "isFriend": user["isFriend"],
                "type": ("follower", user["shortID"])
            })
    # 在这里添加用户自己的数据
    getSongData = getSong(userProfile, song_id, justDATA=True)
    for level in getSongData[0]["songData"]:
        if "Alpha" in level: continue
        result[level].append({
            "playerName": getSongData[1]["playerName"],
            "avatarID": getSongData[1]["playerAvatar"],
            "songScore": getSongData[0]["songData"][level]["songData"]["score"],
            "isFriend": True,
            "type": ("user", "114514")
        })
    # 处理这些数据
    for level in result:
        result[level] = sorted(result[level], key=lambda x: x['songScore'], reverse=True)
        for i in range(len(result[level])):
            result[level][i]["rank"] = i + 1
        if vicinity is not None:
            if vicinity == "user":
                b_index = next((index for index, item in enumerate(result[level]) if item["type"][0] == "user"), None)
                result[level] = result[level][max(b_index - rangeNum, 0):min(b_index + rangeNum + 1, len(result[level]))] if b_index is not None else []
            elif len(vicinity) == 2 and vicinity[0] == "follower":
                b_index = next((index for index, item in enumerate(result[level]) if item["type"][0] == "follower" and item["type"][1].lower() == vicinity[1].lower()), None)
                result[level] = result[level][max(b_index - rangeNum, 0):min(b_index + rangeNum + 1, len(result[level]))] if b_index is not None else []
        
    if justDATA:
        return result
    with open(f"{config.assetsDir}/html/friend_song_score_ranking.html", "r", encoding="utf-8") as f:
        html = f.read()
    htmls = []
    for level in result:
        htmls.append(html.replace("/{{{data}}}/", json.dumps({"data": result[level], "level": level, "songName": getSongData[0]["songName"]}, indent=4, ensure_ascii=False)))
    if justHTML:
        return htmls
    with ThreadPoolExecutor() as executor:
        return list(executor.map(lambda html: utils.render_html_to_jpg((500, 1310), 3, html_data=html, isHTML=True), htmls))

def getAllUserSaveRanking(sorting, vicinity=None, rangeNum=4, maxNum=20, justDATA=False, justHTML=False):
    allUserSaves = getInfo.getAllUsersSave(withObjectID=True)
    result = []
    for user in allUserSaves:
        if sorting == "rating":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": round(user[0][1]["playerRating"], 4),
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno Rating Ranking"
        elif sorting == "fc":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": user[0][1]["playerPlayRecords"]["TotalFc"],
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno FC Ranking"
        elif sorting == "ap":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": user[0][1]["playerPlayRecords"]["TotalAp"],
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno AP Ranking"
        elif sorting == "ap+" or sorting == "app":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": user[0][1]["playerPlayRecords"]["TotalApp"],
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno APP Ranking"
        elif sorting == "miss":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": user[0][1]["playerPlayRecords"]["Miss"],
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno Miss Ranking"
        elif sorting == "good":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": user[0][1]["playerPlayRecords"]["Good"],
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno Good Ranking"
        elif sorting == "perfect":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": user[0][1]["playerPlayRecords"]["Perfect"],
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno Perfect Ranking"
        elif sorting == "perfect+":
            result.append({
                "playerName": user[0][1]["playerName"],
                "sorting": user[0][1]["playerPlayRecords"]["PerfectPlus"],
                "playerAvatar": user[0][1]["playerAvatar"],
                "objectId": user[1]
            })
            title = "Rotaeno PerfectPlus Ranking"
        elif sorting[0] == "song":
            songID = sorting[1]
            songLevel = sorting[2]
            # print(user)
            for song in user[0][0]:
                if song["songID"] == songID and song["songLevelName"] == songLevel and song["score"] != 0:
                    result.append({
                        "playerName": user[0][1]["playerName"],
                        "sorting": song["score"],
                        "playerAvatar": user[0][1]["playerAvatar"],
                        "objectId": user[1]
                    })
            title = f"Rotaeno {sorting[1]} ({songLevel.replace('_', ' ')}) Ranking"
    result = sorted(result, key=lambda x: x["sorting"], reverse=True)
    for i in range(len(result)):
        result[i]["rank"] = i + 1
    if vicinity is not None: # 使用objectID判断(me)
        b_index = next((index for index, item in enumerate(result) if item["objectId"] == vicinity), None)
        result = result[max(b_index - rangeNum, 0):min(b_index + rangeNum + 1, len(result))] if b_index is not None else []
    for i in range(len(result)):
        del result[i]["objectId"]
    result = result[:maxNum]
    
    result = {
        "data": result,
        "title": title
    }
    
    if justDATA:
        return result
    with open(f"{config.assetsDir}/html/user_ranking.html", "r", encoding="utf-8") as f:
        html = f.read()
        html = html.replace("/{{{data}}}/", json.dumps(result, indent=4, ensure_ascii=False))
    if justHTML:
        return html
    return utils.render_html_to_jpg((500, 1100), 1.5, html_data=html, isHTML=True)

def getSongPreviewByteIO(song_id):
    with open(f"{config.assetsDir}/song/preview/{song_id}.wav", "rb") as f: return io.BytesIO(f.read())

updateSave = getInfo._getSave
getLatestSaveTimestamp = getInfo.getLatestSaveTimestamp



import os
import json
import time
from PIL import Image
import matplotlib.pyplot as plt
import numpy as np
from datetime import datetime
import matplotlib.dates as mdates
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from . import database
from . import config

songAlias = database.songAlias
songData = database.songData

def timestamp_to_str(timestamp):
    return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d-%H-%M-%S")

def calculate_level(xp):
    xp_ups = [100, 120, 140, 160, 180, 200, 220, 240, 300, 210]
    xp_ups += [220, 230, 240, 250, 260, 270, 280, 290, 300, 250]
    xp_ups += [260, 270, 280, 290, 300, 310, 320, 330, 340, 350]
    xp_ups += [360, 370, 380, 390, 400, 410, 420, 430, 440, 450]
    xp_ups += [460, 470, 480, 490, 500]
    level = 0
    for xp_up in xp_ups:
        xp -= xp_up
        if xp < 0:
            break
        level += 1
    if xp > 0:
        level += xp / 500
    return level

def getCanUsePlayerAvatar(avatarID):
    return str(avatarID) if os.path.exists(f"./img/avatar/{str(avatarID)}.png") else "player"
def getCanUsePlayerBackground(BackgroundID):
    return str(BackgroundID).replace("background_", "")

def getCanUsePlayerCharacter(characterID):
    return str(characterID).replace("character_", "")

def calculate_song_rating(song_score, rating_real, song_is_cleared):
    next_rating_point = 0.001

    if song_score >= 1010000:
        song_rating = rating_real + 3.7
        next_point_score = 1010000
    elif 1008000 <= song_score < 1010000:
        song_rating = rating_real + 3.4 + (song_score - 1008000) / 10000
        next_point_score = (song_rating + next_rating_point - rating_real - 3.4) * 10000 + 1008000
    elif 1004000 <= song_score < 1008000:
        song_rating = rating_real + 2.4 + (song_score - 1004000) / 4000
        next_point_score = (song_rating + next_rating_point - rating_real - 2.4) * 4000 + 1004000
    elif 1000000 <= song_score < 1004000:
        song_rating = rating_real + 2.0 + (song_score - 1000000) / 10000
        next_point_score = (song_rating + next_rating_point - rating_real - 2.0) * 10000 + 1000000
    elif 980000 <= song_score < 1000000:
        song_rating = rating_real + 1.0 + (song_score - 980000) / 20000
        next_point_score = (song_rating + next_rating_point - rating_real - 1.0) * 20000 + 980000
    elif 950000 <= song_score < 980000:
        song_rating = rating_real + 0.0 + (song_score - 950000) / 30000
        next_point_score = (song_rating + next_rating_point - rating_real - 0.0) * 30000 + 950000
    elif 900000 <= song_score < 950000:
        song_rating = rating_real - 1.0 + (song_score - 900000) / 50000
        next_point_score = (song_rating + next_rating_point - rating_real + 1.0) * 50000 + 900000
    elif 500000 <= song_score < 900000:
        song_rating = rating_real - 5.0 + (song_score - 500000) / 100000
        next_point_score = (song_rating + next_rating_point - rating_real + 5.0) * 100000 + 500000
    else:
        song_rating = 0
        next_point_score = 500000
        
    if song_rating < 0: song_rating = 0
        
    if not song_is_cleared:
        song_rating = min(6, song_rating)
        
    next_point_score -= song_score
    if next_point_score + song_score > 1010000: next_point_score = 1010000 - song_score
    
    return song_rating, next_point_score

def shortIDToUserProfile(shortID):
    with open("users.json", "r", encoding="utf-8") as f:
        users = json.load(f)
    for objectID in os.listdir(config.savesDir):
        for filename in os.listdir(f"{config.savesDir}/{objectID}/userdata"):
            if shortID in filename:
                for userID in users:
                    if users[userID]["objectId"] == objectID: return users[userID]
    return {"server": "", "objectId": "", "sign": "", "session": ""}

def userProfileToShortID(userProfile):
    userProfile["objectId"]
    
    if os.path.exists(f"{config.savesDir}/{userProfile['objectId']}"):
        for filename in os.listdir(f"{config.savesDir}/{userProfile['objectId']}/userdata"):
            return filename.split("-")[1].split(".")[0]
    return ""

def compress_image(image_path, max_size_mb=9.5, quality=95):
    max_size_bytes = max_size_mb * 1024 * 1024
    if os.path.getsize(image_path) > max_size_bytes:
        with Image.open(image_path) as img:
            img.thumbnail((img.width * 0.9, img.height * 0.9), Image.LANCZOS)
            img_format = img.format
            if img_format not in ["JPEG", "JPG", "PNG"]:
                return image_path
            if img.mode in ["RGBA", "P"]:
                img = img.convert("RGB")
            output_path = image_path.rsplit(".", 1)[0] + f".{time.time()}.webp"
            try:
                img.save(output_path, format="WEBP", quality=quality, method=6)
                if os.path.getsize(output_path) <= max_size_bytes:
                    return output_path
            except:
                ...
            output_path = image_path.rsplit(".", 1)[0] + f".{time.time()}.jpg"
            img.save(output_path, format="JPEG", quality=quality, optimize=True)
    else:
        return image_path
    if os.path.getsize(output_path) > max_size_bytes:
        return compress_image(output_path, max_size_mb=max_size_mb, quality=quality-20)
    else:
        return output_path

def render_html_to_jpg(window_size, sleep_time, html_path=None, isHTML=False, html_data=None):    
    options = Options()
    options.add_argument("--headless")
    options.add_argument("--disable-gpu")
    options.add_argument("--allow-file-access-from-files")
    options.add_argument("--enable-local-file-accesses")
    options.add_argument("--no-sandbox")

    driver = webdriver.Chrome(options=options)
    driver.set_window_size(window_size[0], window_size[1])
    
    driver.set_page_load_timeout(30)
    driver.implicitly_wait(10)
    
    if isHTML:
        html_path = f"{config.tmpDir}/{time.time()}.html"
        with open(html_path, "w", encoding="utf-8") as f:
            f.write(html_data)
        
    abs_html_path = os.path.abspath(html_path)
    file_url = f"file://{abs_html_path}"

    driver.get(file_url)
    time.sleep(sleep_time)
    
    total_height = driver.execute_script("return document.body.scrollHeight") + 200
    driver.set_window_size(window_size[0], total_height)
    driver.execute_script("document.body.style.overflow = 'hidden';")

    screenshot_path = f"{config.tmpDir}/{time.time()}.png"
    driver.save_screenshot(screenshot_path)
    driver.quit()
    
    return compress_image(screenshot_path, 9.5)

def draw_bar_chart(data, labels, save_path, title):
    plt.figure(figsize=(12, 6))
    plt.rcParams["font.sans-serif"] = ["SimHei"]
    plt.rcParams["axes.unicode_minus"] = False
    x = np.arange(len(labels))
    plt.bar(x, data, color='cornflowerblue', alpha=0.8, label="Rating")
    p = np.polyfit(x, data, 1)
    line_fit = np.polyval(p, x)
    plt.plot(x, line_fit, color='green', linestyle='--', linewidth=2, label="趋势")
    line_data = [v / 2 for v in data]
    plt.plot(x, line_data, marker='o', color='red', linestyle='-', linewidth=2, markersize=5, label="折线")
    for i, v in enumerate(data):
        plt.text(i, v + 0.2, f"{v:.3f}", ha='center', fontsize=8, rotation=90)
    for i, v in enumerate(line_data):
        plt.text(i, v, f" {v * 2:.3f}", ha='center', fontsize=8, color='red', verticalalignment='bottom', rotation=90)
    plt.xticks(x, labels, rotation=45, ha='right', fontsize=8)
    plt.ylabel("Rating")
    plt.title(title)
    plt.legend()
    plt.tight_layout()
    plt.savefig(save_path)
    plt.close()
    return save_path

def draw_line_datatime(data, save_path, title, y_title):
    dates = [datetime.fromtimestamp(ts) for ts, _ in data]
    ys = [y for _, y in data]
    plt.figure(figsize=(15, 5))
    plt.plot(dates, ys, marker='o', linestyle='-', color='b')
    plt.gca().xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
    plt.gca().xaxis.set_major_locator(mdates.DayLocator())
    plt.gcf().autofmt_xdate()
    plt.title(title)
    plt.xlabel('Date and Time')
    plt.ylabel(y_title)
    plt.grid(True)
    plt.tight_layout()
    plt.savefig(save_path)
    plt.close()
    return save_path


def sorted_json_files(folder_path, reverse=True, needTimestamp=False):
    # reverse=True 新的文件在前
    # reverse=False 旧的文件在前
    files = [f for f in os.listdir(folder_path) if f.endswith(".json")]
    sorted_files = sorted(files, key=lambda f: float(f.split("-")[0] if "-" in f else f.split(".")[0]), reverse=reverse)
    if not needTimestamp:
        return [f"{folder_path}/{file}" for file in sorted_files]
    k = []
    for file in sorted_files:
        k.append([float(file.split(".")[0]), f"{folder_path}/{file}"])
    return k

def safe_open(file_path, mode="r", *args, **kwargs):
    folder = os.path.dirname(file_path)
    if folder and not os.path.exists(folder):
        os.makedirs(folder)
    return open(file_path, mode, *args, **kwargs)



from . import config
from . import utils
from . import database
from . import getInfo
from . import getProcessedData

__all__ = [
    "config",
    "utils",
    "database",
    "getInfo",
    "getProcessedData"
]



from . import config
from . import getProcessedData
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import json
import os

# 定义神经网络模型
class ScorePredictor(nn.Module):
    def __init__(self):
        super(ScorePredictor, self).__init__()
        self.model = nn.Sequential(
            nn.Linear(2, 32),
            nn.ReLU(),
            nn.Linear(32, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, 1)
        )
        self.is_trained = False
    
    def forward(self, x):
        return self.model(x)
    
    def save_model(self, path):
        if not self.is_trained:
            raise ValueError("模型尚未训练，无法保存")
        torch.save(self.state_dict(), path)
    
    def load_model(self, path):
        if not os.path.exists(path):
            raise FileNotFoundError(f"模型文件 {path} 不存在")
        self.load_state_dict(torch.load(path))
        self.is_trained = True
    
    def predict(self, song_level, player_rating):
        if not self.is_trained:
            raise ValueError("模型尚未训练，无法进行预测")
        
        self.eval()
        with torch.no_grad():
            x = torch.FloatTensor([[song_level, player_rating]])
            score = self.model(x)
            return score.item()

def load_real_data():
    with open('AIdata.json', 'r') as f:
        data = json.load(f)
    
    scores = np.array(data[0])
    player_ratings = np.array(data[1])
    song_levels = np.array(data[2])
    
    # 过滤掉无效的分数（小于0或大于20.5的分数）
    valid_indices = (scores >= 0) & (scores <= 20.5)
    scores = scores[valid_indices]
    player_ratings = player_ratings[valid_indices]
    song_levels = song_levels[valid_indices]
    
    # 为高分数据添加随机噪声
    high_scores_mask = scores > 20.0
    noise = np.random.uniform(0.01, 0.05, size=scores[high_scores_mask].shape)
    scores[high_scores_mask] -= noise
    
    # 归一化分数到20.5满分
    max_score = 20.5
    scores = np.minimum(scores, max_score)
    
    # 确保数据维度正确
    scores = scores.reshape(-1, 1)
    player_ratings = player_ratings.reshape(-1, 1)
    song_levels = song_levels.reshape(-1, 1)
     
    return song_levels, player_ratings, scores

# 训练模型
def train_model(model, n_epochs=10000):
    criterion = nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001, weight_decay=0.01)
    
    song_levels, player_levels, scores = load_real_data()
    
    X = torch.FloatTensor(np.hstack([song_levels, player_levels]))
    y = torch.FloatTensor(scores)
    
    if len(X.shape) == 1:
        X = X.view(-1, 2)
    if len(y.shape) == 1:
        y = y.view(-1, 1)
    
    model.train()
    for epoch in range(n_epochs):
        outputs = model(X)
        loss = criterion(outputs, y)
        
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        
        if (epoch + 1) % 10 == 0:
            print(f'Epoch [{epoch+1}/{n_epochs}], Loss: {loss.item():.4f}')
    
    model.is_trained = True


def score(userProfile, level):
    model = ScorePredictor()
    model.load_model(config.scorePredictorModelPath)
    rating = getProcessedData.getPlayerInfo(userProfile, justDATA=True)["playerRating"]
    return int(model.predict(level, rating) * 50000)


def main():
    model = ScorePredictor()
    
    print("开始训练模型...")
    train_model(model)
    
    model.save_model('score_predictor.pth')
    print("模型已保存到 score_predictor.pth")
    
    new_model = ScorePredictor()
    new_model.load_model('score_predictor.pth')
    
    print("\n测试模型...")
    song_level = 14.2
    player_rating = 17.42
    predicted_score = new_model.predict(song_level, player_rating) * 50000
    print(f"歌曲难度: {song_level}, 玩家等级: {player_rating}")
    print(f"预测分数: {predicted_score:.2f}")

if __name__ == "__main__":
    main()

# AIdata.json
# [
#   [score / 50000 ...],
#   [playerRating ...]
#   [songLevel ...]
# ]



import sqlite3
from thefuzz import fuzz
from . import config

class songAlias:
    def __init__(self, database_path):
        self.database_path = database_path
        self.initDatabase()
        self.conn = sqlite3.connect(self.database_path, check_same_thread=False)
    
    def initDatabase(self):
        conn = sqlite3.connect(self.database_path)
        cursor = conn.cursor()
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS song_alias (
                alias TEXT PRIMARY KEY,
                id TEXT
            )
        """)
        conn.commit()
        conn.close()
    
    def getSongID(self, songAlias):
        cursor = self.conn.cursor()
        cursor.execute("SELECT alias, id FROM song_alias")
        records = cursor.fetchall()
        
        res = {}
        if records:
            aliases = [record[0] for record in records]
            match = []
            for alias in aliases:
                match.append((alias, fuzz.token_set_ratio(songAlias, alias)))
            if match:
                for i in match:
                    if i[1] >= 80:
                        cursor.execute("SELECT id FROM song_alias WHERE alias = ?", (i[0],))
                        result = cursor.fetchone()
                        if result is None: continue
                        res[i[0]] = result[0]
            if len(res) > 1:
                for i in res:
                    if i.lower() == songAlias.lower(): return {i: res[i]}
        return res
    
    def addSongAlias(self, songAlias, songID):
        try:
            self.conn.cursor().execute("INSERT INTO song_alias (alias, id) VALUES (?, ?)", (songAlias, songID))
            self.conn.commit()
        except sqlite3.IntegrityError:
            print(f"Error: The alias '{songAlias}' already exists.")
    
    def delSongAlias(self, songAlias):
        self.conn.cursor().execute("DELETE FROM song_alias WHERE alias = ?", (songAlias,))
        self.conn.commit()

class songData:
    def __init__(self, database_path):
        self.database_path = database_path
        self.initDatabase()
        self.conn = sqlite3.connect(self.database_path, check_same_thread=False)
    
    def initDatabase(self):
        conn = sqlite3.connect(self.database_path)
        cursor = conn.cursor()
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS songs (
                id TEXT PRIMARY KEY,
                title TEXT NOT NULL,
                artist TEXT NOT NULL
            )
        """)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS song_levels (
                id TEXT PRIMARY KEY,
                I FLOAT NOT NULL,
                II FLOAT NOT NULL,
                III FLOAT NOT NULL,
                IV FLOAT NOT NULL,
                IV_Alpha FLOAT NOT NULL
            )
        """)
        conn.commit()
        conn.close()
    
    def addSong(self, songID, songTitle, songArtist, songLevels):
        try:
            self.conn.cursor().execute("INSERT INTO songs (id, title, artist) VALUES (?, ?, ?)", (songID, songTitle, songArtist))
            self.conn.cursor().execute("INSERT INTO song_levels (id, I, II, III, IV, IV_Alpha) VALUES (?, ?, ?, ?, ?, ?)",
                                       (songID, songLevels["I"], songLevels["II"], songLevels["III"], songLevels["IV"], songLevels["IV_Alpha"] if "IV_Alpha" in songLevels else 0))
            self.conn.commit()
        except sqlite3.IntegrityError:
            print(f"Error: The song '{songID}' already exists.")
    
    def delSong(self, songID):
        self.conn.cursor().execute("DELETE FROM songs WHERE id = ?", (songID,))
        self.conn.commit()
    
    def getSong(self, songID):
        res = {}
        cursor = self.conn.cursor()
        cursor.execute("SELECT title, artist FROM songs WHERE id = ?", (songID,))
        result = cursor.fetchone()
        if result is None: return res
        res["id"] = songID
        res["title"] = result[0]
        res["artist"] = result[1]
        cursor.execute("SELECT I, II, III, IV, IV_Alpha FROM song_levels WHERE id = ?", (songID,))
        result = cursor.fetchone()
        if result is None: return res
        res["levels"] = [
            result[0], # I
            result[1], # II
            result[2], # III
            result[3], # IV
            result[4] # IV_Alpha
        ]
        res["__levels"] = {
            "I": result[0], # I
            "II": result[1], # II
            "III": result[2], # III
            "IV": result[3], # IV
            "IV_Alpha": result[4] # IV_Alpha
        }
        return res
    
    def getSongsRatingRealRange(self, songRatingReal1, songRatingReal2):
        cursor = self.conn.cursor()
        cursor.execute('''
            SELECT id, I, II, III, IV, IV_Alpha
            FROM song_levels
            WHERE (I BETWEEN ? AND ?)
            OR (II BETWEEN ? AND ?)
            OR (III BETWEEN ? AND ?)
            OR (IV BETWEEN ? AND ?)
            OR (IV_Alpha BETWEEN ? AND ?)
        ''', (songRatingReal1, songRatingReal2, songRatingReal1, songRatingReal2, songRatingReal1, songRatingReal2, songRatingReal1, songRatingReal2, songRatingReal1, songRatingReal2))
        # return cursor.fetchall()
        res = []
        level_map = {1: "I", 2: "II", 3: "III", 4: "IV", 5: "IV_Alpha"}
        for row in cursor.fetchall():
            tmp = {}
            for level in level_map:
                if row[level] >= songRatingReal1 and row[level] <= songRatingReal2 and row[level] != 0:
                    tmp["id"] = row[0]
                    try:
                        tmp["levels"].append(level_map[level])
                    except:
                        tmp["levels"] = [level_map[level]]
            if tmp != {}: res.append(tmp)
        return res

songAlias = songAlias(config.songAliasDBPath)
songData = songData(config.songDataDBPath)
